use itertools::Itertools;
use rand::prelude::*;
use rayon::prelude::*;
use std::collections::HashMap;

fn find_rotate_dp(
    ring: &[u8],
    key: &[u8],
    current_ring_position: usize,
    current_key_position: usize,
    memo: &mut HashMap<(usize, usize), i32>,
) -> i32 {
    if current_key_position == key.len() {
        0
    } else {
        if let Some(cached) = memo.get(&(current_ring_position, current_key_position)) {
            return *cached;
        }
        let value = if ring[current_ring_position] == key[current_key_position] {
            1 + find_rotate_dp(
                ring,
                key,
                current_ring_position,
                current_key_position + 1,
                memo,
            )
        } else {
            // we can either go left or right
            let left_chars = ring[0..current_ring_position]
                .iter()
                .rev()
                .chain(ring[(current_ring_position + 1)..].iter().rev());
            let right_chars = ring[(current_ring_position + 1)..]
                .iter()
                .chain(ring[0..current_ring_position].iter());
            let target = key[current_key_position];
            let left_movements = left_chars.take_while(|&c| *c != target).count() + 1;
            let right_movements = right_chars.take_while(|&c| *c != target).count() + 1;
            let left_pos = (current_ring_position + ring.len() - left_movements) % ring.len();
            let right_pos = (current_ring_position + right_movements) % ring.len();
            assert!(ring[left_pos] == target);
            assert!(ring[right_pos] == target);
            let left_value = find_rotate_dp(ring, key, left_pos, current_key_position, memo)
                + left_movements as i32;
            let right_value = find_rotate_dp(ring, key, right_pos, current_key_position, memo)
                + right_movements as i32;
            left_value.min(right_value)
        };
        memo.insert((current_ring_position, current_key_position), value);
        value
    }
}

fn find_rotate_steps(ring: &str, key: &str) -> i32 {
    let mut memo = HashMap::new();
    find_rotate_dp(ring.as_bytes(), key.as_bytes(), 0, 0, &mut memo)
}

fn find_rotate_steps_par(ring: &str, key: &str) -> i32 {
    // let's figure out where each unique char is
    let unique_chars = ring
        .as_bytes()
        .par_iter()
        .enumerate()
        .fold(HashMap::new, |mut h: HashMap<u8, (bool, usize)>, (i, c)| {
            match h.entry(*c) {
                std::collections::hash_map::Entry::Occupied(mut o) => {
                    o.get_mut().0 = false;
                }
                std::collections::hash_map::Entry::Vacant(v) => {
                    v.insert((true, i));
                }
            }
            h
        })
        .reduce(HashMap::new, |mut h1, h2| {
            h2.into_iter().for_each(|(c, (unique, pos))| {
                if let Some(existing_c) = h1.get_mut(&c) {
                    existing_c.0 = false;
                } else {
                    h1.insert(c, (unique, pos));
                }
            });
            h1
        })
        .into_iter()
        .filter_map(|(c, (unique, pos))| if unique { Some((c, pos)) } else { None })
        .collect::<HashMap<_, _>>();

    rayon::iter::split((0, key.as_bytes()), |(start_index, key_part)| {
        // look for a unique char around the midpoint
        // if we find one we split around it
        let midpoint = key_part.len() / 2;
        let unique_char_index = (midpoint..key_part.len() - 1)
            .interleave((1..midpoint).rev())
            .filter(|&i| unique_chars.contains_key(&key_part[i]))
            .next();
        if let Some(split_index) = unique_char_index {
            // we want the midpoint on the left side
            let (left_key, right_key) = key_part.split_at(split_index + 1);
            let midchar = *left_key.last().unwrap();
            let midchar_pos = unique_chars[&midchar];
            ((start_index, left_key), Some((midchar_pos, right_key)))
        } else {
            ((start_index, key_part), None)
        }
    })
    .map(|(ring_position, key)| {
        let mut memo = HashMap::new();
        find_rotate_dp(ring.as_bytes(), key, ring_position, 0, &mut memo)
    })
    .sum()
}

fn random_input(ring_size: usize, key_size: usize) -> (String, String) {
    fn random_letter() -> char {
        char::from_u32('a' as u32 + rand::random::<u32>() % 26).unwrap()
    }
    let ring: String = std::iter::repeat_with(random_letter)
        .take(ring_size)
        .collect();
    let ring_letters = ring.chars().unique().collect::<Vec<char>>();
    let mut rng = rand::thread_rng();
    let key = std::iter::repeat_with(|| ring_letters.choose(&mut rng).unwrap())
        .take(key_size)
        .collect();
    (ring, key)
}

fn main() {
    let (ring, key) = random_input(100, 1000);
    let start = std::time::Instant::now();
    let rseq = find_rotate_steps(&ring, &key);
    println!("seq took {:?}", start.elapsed());

    let start = std::time::Instant::now();
    let rpar = find_rotate_steps_par(&ring, &key);
    println!("par took {:?}", start.elapsed());

    assert_eq!(rseq, rpar);
}
